/*
 *  Copyright (c) by Abramo Bagnara
 *  <abbagnara@racine.ra.it>
 *  Routines for control of ESS ES18xx chip
 *
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/* GENERAL NOTES:
 *
 * BUGS:
 * - There are pops (we can't delay in trigger function, cause midlevel 
 *   often need to trigger down and then up very quickly).
 *   Any ideas?
 */

/*
 * ES1868  NOTES:
 * - The chip has one half duplex pcm (with very limited full duplex support).
 *
 * - Duplex stereophonic sound is impossible.
 * - Record and playback must share the same frequency rate.
 *
 * - The driver use dma2 for playback and dma1 for record.
 */

/*
 * ES1869 NOTES:
 *
 * - there are a first full duplex pcm and a second playback only pcm
 *   (incompatible with first pcm record)
 * 
 * - there is support for the record volume and ESS Spatializer 3D effect.
 *
 * - contrarily to some pages in DS_1869.PDF the rates can be set
 *   independently.
 *
 * BUGS:
 *
 * - There is a major trouble I noted:
 *
 *   using both channel for playback stereo 16 bit samples at 44100 Hz
 *   the second pcm (Audio1) DMA slows down irregularly and sound is garbled.
 *   
 *   The same happens using Audio1 for recording.
 *
 *   The Windows driver does not suffer of this (although it use Audio1
 *   only for recording). I'm unable to discover why.
 *
 *   --- UPDATE ---
 *   This happened when I used DMA 1 for Audio1 and DMA 0 for Audio2.
 *   Using DMA 0 for Audio1 and DMA1 for Audio2 this trouble is magically
 *   disappeared!!!
 *   Now alsa-lib/test/latency gives a 2000 us latency for fragment size 128.
 *   Formerly the underruns was always present for all fragment sizes.
 *
 */

/*
 * ES1878 NOTES:
 * ES1879 NOTES:
 * ES1887 NOTES:
 * ES1888 NOTES:
 *
 * ------- WARNING --------
 * THESE CHIPS ARE NEVER BEEN TESTED.
 * THEIR PnP ID MUST BE ADDED TO audiodrive18xx.c
 * THEIR CAPABILITIES MUST BE FIXED IN snd_es18xx_probe.
 * ------- WARNING --------
 */

#define __SND_OSS_COMPAT__
#define SND_MAIN_OBJECT_FILE
#include "driver.h"
#include "es18xx.h"


#define DAC1 0x01
#define ADC1 0x02
#define DAC2 0x04
#define MILLISECOND 10000

static int snd_es18xx_dsp_command(es18xx_t * codec, unsigned char val)
{
        int i;

        for(i = MILLISECOND; i; i--)
                if ((inb(codec->port + 0x0C) & 0x80) == 0) {
                        outb(val, codec->port + 0x0C);
                        return 1;
                }
        snd_printk("snd_es18xx_dsp_command: timeout (0x%x)\n", val);
        return 0;
}

static int snd_es18xx_dsp_get_byte(es18xx_t * codec)
{
        int i;

        for(i = MILLISECOND/10; i; i--)
                if (inb(codec->port + 0x0C) & 0x40)
                        return inb(codec->port + 0x0A);
        snd_printk("es18xx get byte failed: 0x%x = 0x%x!!!\n", codec->port + 0x0A, inb(codec->port + 0x0A));
        return -ENODEV;
}

#undef REG_DEBUG

static int snd_es18xx_write(es18xx_t * codec,
			    unsigned char reg, unsigned char data)
{
#ifdef REG_DEBUG
	snd_printk("Reg %02x := %02x\n", reg, data);
#endif
        if (!snd_es18xx_dsp_command(codec, reg))
                return 0;

        return snd_es18xx_dsp_command(codec, data);
}

static int snd_es18xx_read(es18xx_t *codec, unsigned char reg)
{
	int data;
        /* Read a byte from an extended mode register of ES18xx */
        if (!snd_es18xx_dsp_command(codec, 0xC0))
                return -1;
        if (!snd_es18xx_dsp_command(codec, reg))
                return -1;
	data = snd_es18xx_dsp_get_byte(codec);
#ifdef REG_DEBUG
	snd_printk("Reg %02x == %02x\n", reg, data);
#endif
	return data;
}

static int snd_es18xx_bits(es18xx_t *codec, unsigned char reg,
			   unsigned char set, unsigned char unset)
{
        unsigned char c = snd_es18xx_read(codec, reg);

        if (!c)
                return -1;
        return snd_es18xx_write(codec, reg, (c & ~unset) | set);
}

inline void snd_es18xx_mixer_write(es18xx_t * codec,
			    unsigned char reg, unsigned char data)
{
#ifdef REG_DEBUG
	snd_printk("Mixer reg %02x set to %02x\n", reg, data);
#endif
        outb(reg, codec->port + 0x04);
        outb(data, codec->port + 0x05);
}

inline unsigned char snd_es18xx_mixer_read(es18xx_t * codec, unsigned char reg)
{
	unsigned char data;
        outb(reg, codec->port + 0x04);
	data = inb(codec->port + 0x05);
#ifdef REG_DEBUG
	snd_printk("Mixer reg %02x now is %02x\n", reg, data);
#endif
        return data;
}

static inline int snd_es18xx_mixer_bits(es18xx_t *codec, unsigned char reg,
					unsigned char set, unsigned char unset)
{
	unsigned char old;
	unsigned char new;
        outb(reg, codec->port + 0x04);
	old = inb(codec->port + 0x05);
	new = (old & ~unset) | set;
	outb(new, codec->port + 0x05);
#ifdef REG_DEBUG
	snd_printk("Mixer reg %02x was %02x now is %02x\n", reg, old, new);
#endif
	return new;
}


static int snd_es18xx_reset(es18xx_t * codec)
{
        int i;

        outb(0x03, codec->port + 0x06);
        inb(codec->port + 0x06);
        outb(0x00, codec->port + 0x06);
        for(i = 0; i < MILLISECOND && !(inb(codec->port + 0x0E) & 0x80); i++);
        if (inb(codec->port + 0x0A) != 0xAA) {
                snd_printk("es18xx_reset at 0x%x: failed!!!\n", codec->port);
                return -ENODEV;
        }
        /* enable extended mode */
        snd_es18xx_dsp_command(codec, 0xC6);
	/* Reset mixer registers */
	snd_es18xx_mixer_write(codec, 0x00, 0x00);
        /* Enable Audio 1 IRQ */
        snd_es18xx_write(codec, 0xB1, 0x50);
	/* Enable Audio 2 IRQ */
        snd_es18xx_mixer_write(codec, 0x7A, 0x40);
        /* Enable Audio 1 DMA */
        snd_es18xx_write(codec, 0xB2, 0x50);
        /* Audio 1 DMA demand mode (4 bytes/request) */
        snd_es18xx_write(codec, 0xB9, 2);
	/* Enable MPU and hardware volume interrupt */
	snd_es18xx_mixer_write(codec, 0x64, 0x42);
	if (codec->caps & ES18XX_PCM2) {
		/* Change behaviour of register A1
		   4x oversampling
		   2nd channel DAC asynchronous */
		snd_es18xx_mixer_write(codec, 0x71, 0x32);
	}
	else {
		/* Enable DMA FIFO */
		snd_es18xx_write(codec, 0xB7, 0x80);
	}
	/* Record source is record mixer */
	if (codec->caps & ES18XX_RECMIX)
		snd_es18xx_mixer_write(codec, 0x1c, 0x05);
	if (codec->caps & ES18XX_3D) {
		/* Set 3D parameter to recommended values */
		snd_es18xx_mixer_write(codec, 0x54, 0x8f);
		snd_es18xx_mixer_write(codec, 0x56, 0x95);
		snd_es18xx_mixer_write(codec, 0x58, 0x94);
		snd_es18xx_mixer_write(codec, 0x5a, 0x80);
	}
        return 0;
}

static int snd_es18xx_reset_fifo(es18xx_t * codec)
{
        outb(0x02, codec->port + 0x06);
        inb(codec->port + 0x06);
        outb(0x00, codec->port + 0x06);
	return 0;
}

static int snd_es18xx_probe(es18xx_t *codec)
{
        unsigned long flags;
        unsigned short port;

        port = codec->port;

        /* I am not sure if this is nessesary */
        snd_spin_lock(codec, reg, &flags);

        /*
         *  identifying the ES18xx
         */

        outb(0x40, port + 0x04);
        codec->version =  inb(port + 0x05) << 8;
        codec->version += inb(port + 0x05);
        codec->ctrl_port =  inb(port + 0x05) << 8;
        codec->ctrl_port += inb(port + 0x05);
	switch (codec->version) {
	case 0x1868:
		codec->caps = ES18XX_DUPLEX_MONO | ES18XX_DUPLEX_SAME;
		break;
	case 0x1869:
		codec->caps = ES18XX_PCM2 | ES18XX_3D | ES18XX_RECMIX | ES18XX_NEW_RATE | ES18XX_AUXB | ES18XX_SPEAKER;
		break;
	case 0x1878:
		/* FIXME */
		codec->caps = ES18XX_DUPLEX_MONO | ES18XX_DUPLEX_SAME;
		break;
	case 0x1879:
		/* FIXME */
		codec->caps = ES18XX_PCM2 | ES18XX_3D | ES18XX_RECMIX | ES18XX_NEW_RATE | ES18XX_AUXB | ES18XX_SPEAKER;
		break;
	case 0x1887:
		/* FIXME */
		codec->caps = ES18XX_PCM2 | ES18XX_RECMIX | ES18XX_NEW_RATE | ES18XX_AUXB | ES18XX_SPEAKER;
		break;
	case 0x1888:
		/* FIXME */
		codec->caps = ES18XX_PCM2 | ES18XX_RECMIX | ES18XX_NEW_RATE | ES18XX_AUXB | ES18XX_SPEAKER;
		break;
	default:
                snd_printk("[0x%x] ESS: unknown AudioDrive chip ES%x\n",
                           port, codec->version);
                snd_spin_unlock(codec, reg, &flags);
                return -ENODEV;
        }

        /*
         *  reset the ES18xx
         */

        if (snd_es18xx_reset(codec) < 0) {
                snd_printk("ESS: [0x%x] reset failed... 0x%x\n",
                            port, inb(port + 0x0A));
                snd_spin_unlock(codec, reg, &flags);
                return -ENODEV;
        }

        snd_spin_unlock(codec, reg, &flags);

        snd_printd("ESS: [0x%x] found.. version = 0x%x, ctrl port = 0x%x\n", codec->port, codec->version, codec->ctrl_port);

        return 0;
}

static int snd_es18xx_set_rate(es18xx_t *codec, int rate, int mode, int set)
{
        int div0, div1, diff0, diff1, rate0, rate1;
        unsigned char bits;

	if (codec->caps & ES18XX_NEW_RATE) {
		/* let us calculate the best clock choice */
		div0 = (793800 + (rate / 2)) / rate;
		rate0 = 793800 / div0;
		diff0 = (rate0 > rate) ? (rate0 - rate) : (rate - rate0);
		
		div1 = (768000 + (rate / 2)) / rate;
		rate1 = 768000 / div1;
		diff1 = (rate1 > rate) ? (rate1 - rate) : (rate - rate1);
		
		if (diff0 < diff1) {
			bits = 128 - div0;
			rate = rate0;
		}
		else {
			bits = 256 - div1;
			rate = rate1;
		}
	}
	else {
		if (rate > 22000) {
			div0 = (795444 + (rate / 2)) / rate;
			rate = 795444 / div0;
			bits = 256 - div0;

		}
		else {
			div0 = (397722 + (rate / 2)) / rate;
			rate = 397722 / div0;
			bits = 128 - div0;
		}
	}

	if (set) {
		/* set filter register */
		div0 = 256 - 7160000*20/(8*82*rate);
		
		if ((codec->caps & ES18XX_PCM2) && mode == DAC2) {
			snd_es18xx_mixer_write(codec, 0x70, bits);
			snd_es18xx_mixer_write(codec, 0x72, div0);
		}
		else {
			snd_es18xx_write(codec, 0xA1, bits);
			snd_es18xx_write(codec, 0xA2, div0);
		}
	}
	return rate;
}

static int snd_es18xx_playback_ioctl(snd_pcm1_t * pcm1,
				     unsigned int cmd, unsigned long *arg)
{
        es18xx_t *codec;
        codec = (es18xx_t *) pcm1->private_data;
        switch(cmd) {
        case SND_PCM1_IOCTL_RATE:
                pcm1->playback.real_rate = snd_es18xx_set_rate(codec, pcm1->playback.rate, DAC2, 0);
#if 0
		if ((codec->caps & ES18XX_DUPLEX_SAME) &&
		    pcm1->flags & SND_PCM1_LFLG_RECORD) {
			if (pcm1->playback.real_rate != pcm1->record.real_rate)
				return -EBUSY;
		}
#endif
                return 0;
	case SND_PCM1_IOCTL_VOICES:
		if ((codec->caps & ES18XX_DUPLEX_MONO) &&
		    pcm1->flags & SND_PCM1_LFLG_RECORD &&
		    (pcm1->playback.voices != 1 ||
		     pcm1->record.voices != 1))
			return -EBUSY;
		else
			return 0;
        }
        return -ENXIO;
}

static int snd_es18xx_playback2_ioctl(snd_pcm1_t * pcm1,
				     unsigned int cmd, unsigned long *arg)
{
        es18xx_t *codec;
        codec = (es18xx_t *) pcm1->private_data;
        switch(cmd) {
        case SND_PCM1_IOCTL_RATE:
                pcm1->playback.real_rate = snd_es18xx_set_rate(codec, pcm1->playback.rate, DAC1, 0);
                return 0;
        }
        return -ENXIO;
}

static int snd_es18xx_record_ioctl(snd_pcm1_t * pcm1,
				   unsigned int cmd, unsigned long *arg)
{
        es18xx_t *codec;
        codec = (es18xx_t *) pcm1->private_data;
        switch(cmd) {
        case SND_PCM1_IOCTL_RATE:
                pcm1->record.real_rate = snd_es18xx_set_rate(codec, pcm1->record.rate, ADC1, 0);
#if 0
		if ((codec->caps & ES18XX_DUPLEX_SAME) &&
		    pcm1->flags & SND_PCM1_LFLG_PLAY) {
			if (pcm1->playback.real_rate != pcm1->record.real_rate)
				return -EBUSY;
		}
#endif
                return 0;
	case SND_PCM1_IOCTL_VOICES:
		if ((codec->caps & ES18XX_DUPLEX_MONO) &&
		    pcm1->flags & SND_PCM1_LFLG_PLAY &&
		    (pcm1->playback.voices != 1 ||
		     pcm1->record.voices != 1))
			return -EBUSY;
		else
			return 0;
        }
        return -ENXIO;
}

static void snd_es18xx_playback_prepare(snd_pcm1_t *pcm1,
					unsigned char *buffer,
					unsigned int size,
					unsigned int offset,
					unsigned int count)
{
        unsigned long flags;
        es18xx_t *codec;
        snd_pcm1_channel_t *pchn;

        pchn = &pcm1->playback;

        codec = (es18xx_t *) pcm1->private_data;

        snd_spin_lock(codec, mixer, &flags);

        snd_es18xx_set_rate(codec, pchn->rate, DAC2, 1);

        /* Transfer Count Reload */
        count = 0x10000 - count;
        snd_es18xx_mixer_write(codec, 0x74, count & 0xff);
        snd_es18xx_mixer_write(codec, 0x76, count >> 8);

	/* Set format */
        snd_es18xx_mixer_bits(codec, 0x7A, 
			      ((pchn->voices == 1) ? 0x00 : 0x02) |
			      ((pchn->mode & SND_PCM1_MODE_16) ? 0x01 : 0x00) |
			      ((pchn->mode & SND_PCM1_MODE_U) ? 0x00 : 0x04),
			      0x07);


        snd_spin_unlock(codec, mixer, &flags);

        /* Set DMA controller */
        snd_dma_program(codec->dma2, buffer, size, DMA_MODE_WRITE | DMA_MODE_AUTOINIT);

#if 0
	snd_printd("Playback settings = (%d,%d,%s,%s)\n",
		pchn->rate, 
		(pchn->mode & SND_PCM1_MODE_16) ? 16 : 8,
		(pchn->voices==1) ? "mono":"stereo",
		(pchn->mode & SND_PCM1_MODE_U)  ? "unsigned":"signed");
#endif
}

static void snd_es18xx_playback_trigger(snd_pcm1_t * pcm1, int up)
{
        es18xx_t *codec;
        unsigned long flags;

	codec = (es18xx_t *) pcm1->private_data;
        if (up) {
		if (codec->active & DAC2)
			return;
		codec->active |= DAC2;
	}
	else {
		if (!(codec->active & DAC2))
			return;
		codec->active &= ~DAC2;
	}

	snd_spin_lock(codec, reg, &flags);
	if (up) {	
                /* Start DMA */
                snd_es18xx_mixer_write(codec, 0x78, 0x93);
#if 0
		/* Avoid pops */
                snd_delay(88200/pcm1->playback.real_rate);
#endif
		if (codec->caps & ES18XX_PCM2)
			/* Restore Audio 2 volume */
			snd_es18xx_mixer_write(codec, 0x7C, codec->audio2_vol);
		else
			/* Enable PCM output */
			snd_es18xx_dsp_command(codec, 0xD1);
        }
        else {
                /* Stop DMA */
                snd_es18xx_mixer_write(codec, 0x78, 0x00);
#if 0
                snd_delay(3);
#endif
		if (codec->caps & ES18XX_PCM2)
			/* Set Audio 2 volume to 0 */
			snd_es18xx_mixer_write(codec, 0x7C, 0);
		else
			/* Disable PCM output */
			snd_es18xx_dsp_command(codec, 0xD3);
        }

        snd_spin_unlock(codec, reg, &flags);
}

static void snd_es18xx_record_prepare(snd_pcm1_t *pcm1,
				      unsigned char *buffer,
				      unsigned int size,
				      unsigned int offset,
				      unsigned int count)
{
        unsigned long flags;
        es18xx_t *codec;
        snd_pcm1_channel_t *pchn;

        pchn = &pcm1->record;
        codec = (es18xx_t *) pcm1->private_data;

        snd_spin_lock(codec, reg, &flags);

	snd_es18xx_reset_fifo(codec);

        /* Set stereo/mono */
        snd_es18xx_bits(codec, 0xA8, pchn->voices == 1 ? 0x02 : 0x01, 0x0b);

        snd_es18xx_set_rate(codec, pchn->rate, ADC1, 1);

        /* Transfer Count Reload */
	count = 0x10000 - count;
	snd_es18xx_write(codec, 0xA4, count & 0xff);
	snd_es18xx_write(codec, 0xA5, count >> 8);

#if 0
	/* Delay 10 milliseconds to allow the analog circuits to settle */
	snd_delay(100);
#endif

        /* Set format */
        snd_es18xx_write(codec, 0xB7, 
                         (pchn->mode & SND_PCM1_MODE_U) ? 0x51 : 0x71);
        snd_es18xx_write(codec, 0xB7, 0x90 |
                         ((pchn->voices == 1) ? 0x40 : 0x08) |
                         ((pchn->mode & SND_PCM1_MODE_16) ? 0x04 : 0x00) |
                         ((pchn->mode & SND_PCM1_MODE_U) ? 0x00 : 0x20));

        snd_spin_unlock(codec, reg, &flags);

        /* Set DMA controler */
        snd_dma_program(codec->dma1, buffer, size, DMA_MODE_READ | DMA_MODE_AUTOINIT);

#if 0
	snd_printd("Record settings = (%d,%d,%s,%s)\n",
			pchn->rate, (pchn->mode & SND_PCM1_MODE_16) ? 16 : 8,
			(pchn->voices==1) ? "mono":"stereo",
			(pchn->mode & SND_PCM1_MODE_U)  ? "unsigned":"signed");
#endif
}

static void snd_es18xx_record_trigger(snd_pcm1_t *pcm1, int up)
{
        es18xx_t *codec;
        unsigned long flags;

	codec = (es18xx_t *) pcm1->private_data;
        if (up) {
		if (codec->active & ADC1)
			return;
		codec->active |= ADC1;
	}
        else {
		if (!(codec->active & ADC1))
			return;
		codec->active &= ~ADC1;
	}

	snd_spin_lock(codec, reg, &flags);
	if (up)
                /* Start DMA */
                snd_es18xx_write(codec, 0xB8, 0x0f);
        else
                /* Stop DMA */
                snd_es18xx_write(codec, 0xB8, 0x00);
        snd_spin_unlock(codec, reg, &flags);
}

static void snd_es18xx_playback2_prepare(snd_pcm1_t *pcm1,
                                         unsigned char *buffer,
                                         unsigned int size,
                                         unsigned int offset,
                                         unsigned int count)
{
        unsigned long flags;
        es18xx_t *codec;
        snd_pcm1_channel_t *pchn;

        pchn = &pcm1->playback;
        codec = (es18xx_t *) pcm1->private_data;

        snd_spin_lock(codec, reg, &flags);

	snd_es18xx_reset_fifo(codec);

        /* Set stereo/mono */
        snd_es18xx_bits(codec, 0xA8, pchn->voices == 1 ? 0x02 : 0x01, 0x03);

        snd_es18xx_set_rate(codec, pchn->rate, DAC1, 1);

        /* Transfer Count Reload */
	count = 0x10000 - count;
	snd_es18xx_write(codec, 0xA4, count & 0xff);
	snd_es18xx_write(codec, 0xA5, count >> 8);

        /* Set format */
        snd_es18xx_write(codec, 0xB6,
                         (pchn->mode & SND_PCM1_MODE_U) ? 0x80 : 0x00);
        snd_es18xx_write(codec, 0xB7, 
                         (pchn->mode & SND_PCM1_MODE_U) ? 0x51 : 0x71);
        snd_es18xx_write(codec, 0xB7, 0x90 |
                         ((pchn->voices == 1) ? 0x40 : 0x08) |
                         ((pchn->mode & SND_PCM1_MODE_16) ? 0x04 : 0x00) |
                         ((pchn->mode & SND_PCM1_MODE_U) ? 0x00 : 0x20));


        snd_spin_unlock(codec, reg, &flags);

        /* Set DMA controler */
        snd_dma_program(codec->dma1, buffer, size, DMA_MODE_WRITE | DMA_MODE_AUTOINIT);

#if 0
	snd_printd("Playback2 settings = (%d,%d,%s,%s)\n",
		pchn->rate, 
		(pchn->mode & SND_PCM1_MODE_16) ? 16 : 8,
		(pchn->voices==1) ? "mono":"stereo",
		(pchn->mode & SND_PCM1_MODE_U)  ? "unsigned":"signed");
#endif
}

static void snd_es18xx_playback2_trigger(snd_pcm1_t *pcm1, int up)
{
        es18xx_t *codec = (es18xx_t *) pcm1->private_data;
        unsigned long flags;


        if (up) {
		if (codec->active & DAC1)
			return;
		codec->active |= DAC1;
        }
        else {
		if (!(codec->active & DAC1))
			return;
		codec->active &= ~DAC1;
	}

	snd_spin_lock(codec, reg, &flags);
	if (up) {
                /* Start DMA */
                snd_es18xx_write(codec, 0xB8, 0x05);
#if 0
		/* Avoid pops */
                snd_delay(88200/pcm1->playback.real_rate);
#endif
                /* Enable Audio 1 */
                snd_es18xx_dsp_command(codec, 0xD1);
        }
        else {
                /* Stop DMA */
                snd_es18xx_write(codec, 0xB8, 0x00);
#if 0
                snd_delay(3);
#endif
                /* Disable Audio 1 */
                snd_es18xx_dsp_command(codec, 0xD3);
        }
        snd_spin_unlock(codec, reg, &flags);
}

void snd_es18xx_interrupt(es18xx_t *codec, unsigned char status)
{
        /* ok.. playback is active */
        if (status & 0x02) {
                if (codec->active & DAC2) {
                        snd_pcm1_t *pcm1=(snd_pcm1_t *) codec->pcm_a->private_data;
			snd_pcm1_channel_t *pchn = &pcm1->playback;
                        pchn->ack(pcm1);
                }
		/* ack interrupt */
                snd_es18xx_mixer_bits(codec, 0x7A, 0x00, 0x80);
        }

        if (status & 0x01) {
                /* ok.. record is active */
                if (codec->active & ADC1) {
                        snd_pcm1_t *pcm1=(snd_pcm1_t *) codec->pcm_a->private_data;
			snd_pcm1_channel_t *pchn = &pcm1->record;
                        pchn->ack(pcm1);
                }
                /* ok.. playback2 is active */
                else if (codec->active & DAC1) {
                        snd_pcm1_t *pcm1=(snd_pcm1_t *) codec->pcm_b->private_data;
			snd_pcm1_channel_t *pchn = &pcm1->playback;
                        pchn->ack(pcm1);
                }
		/* ack interrupt */
		inb(codec->port + 0x0E);
        }

}

static int snd_es18xx_playback_open(snd_pcm1_t * pcm1)
{
        es18xx_t *codec;
        int err;

        codec = (es18xx_t *) pcm1->private_data;
	if ((codec->caps & ES18XX_DUPLEX_MONO) &&
	    (pcm1->flags & SND_PCM1_LFLG_RECORD) && 
	    pcm1->record.voices != 1)
		return -EBUSY;
        if ((err = snd_pcm1_dma_alloc(pcm1, SND_PCM1_PLAYBACK, codec->dma2ptr, "ES18xx (playback)")) < 0)
                return err;
        codec->open |= DAC2;
        return 0;
}

static int snd_es18xx_playback2_open(snd_pcm1_t * pcm1)
{
        es18xx_t *codec;
        int err;

        codec = (es18xx_t *) pcm1->private_data;
        if (codec->open & ADC1)
                return -EBUSY;
        if ((err = snd_pcm1_dma_alloc(pcm1, SND_PCM1_PLAYBACK, codec->dma1ptr, "ES18xx (playback2)")) < 0)
                return err;
        codec->open |= DAC1;
        return 0;
}

static int snd_es18xx_record_open(snd_pcm1_t * pcm1)
{
        es18xx_t *codec;
        int err;

        codec = (es18xx_t *) pcm1->private_data;
        if (codec->open & DAC1)
                return -EBUSY;
	if ((codec->caps & ES18XX_DUPLEX_MONO) &&
	    (pcm1->flags & SND_PCM1_LFLG_PLAY) && 
	    pcm1->playback.voices != 1)
		return -EBUSY;
        if ((err = snd_pcm1_dma_alloc(pcm1, SND_PCM1_RECORD, codec->dma1ptr, "ES18xx (record)")) < 0)
                return err;
        codec->open |= ADC1;
        return 0;
}

static void snd_es18xx_playback_close(snd_pcm1_t * pcm1)
{
        es18xx_t *codec;

        codec = (es18xx_t *) pcm1->private_data;
        snd_pcm1_dma_free(pcm1, SND_PCM1_PLAYBACK, codec->dma2ptr);
        codec->open &= ~DAC2;
}

static void snd_es18xx_playback2_close(snd_pcm1_t * pcm1)
{
        es18xx_t *codec;

        codec = (es18xx_t *) pcm1->private_data;
        snd_pcm1_dma_free(pcm1, SND_PCM1_PLAYBACK, codec->dma1ptr);
        codec->open &= ~DAC1;
}

static void snd_es18xx_record_close(snd_pcm1_t * pcm1)
{
        es18xx_t *codec;

        codec = (es18xx_t *) pcm1->private_data;
        snd_pcm1_dma_free(pcm1, SND_PCM1_RECORD, codec->dma1ptr);
        codec->open &= ~ADC1;
}

static unsigned int snd_es18xx_playback_pointer(snd_pcm1_t * pcm1, unsigned int used_size)
{
        es18xx_t *codec;

        codec = (es18xx_t *) pcm1->private_data;
        if (!(codec->active & DAC2))
                return 0;
        return used_size - snd_dma_residue(codec->dma2);
}

static unsigned int snd_es18xx_playback2_pointer(snd_pcm1_t * pcm1, unsigned int used_size)
{
        es18xx_t *codec;

        codec = (es18xx_t *) pcm1->private_data;
        if (!(codec->active & DAC1))
                return 0;
        return used_size - snd_dma_residue(codec->dma1);
}

static unsigned int snd_es18xx_record_pointer(snd_pcm1_t * pcm1, unsigned int used_size)
{
        es18xx_t *codec;

        codec = (es18xx_t *) pcm1->private_data;
        if (!(codec->active & ADC1))
                return 0;
        return used_size - snd_dma_residue(codec->dma1);
}

static struct snd_stru_pcm1_hardware snd_es18xx_playback =
{
        NULL,			/* private data */
        NULL,			/* private_free */
        SND_PCM1_HW_AUTODMA,	/* flags */
        SND_PCM_FMT_MU_LAW |
        SND_PCM_FMT_U8 | SND_PCM_FMT_S8 | 
        SND_PCM_FMT_S16_LE | SND_PCM_FMT_U16_LE,	/* formats */
        SND_PCM_FMT_U8 | SND_PCM_FMT_S8 | 
        SND_PCM_FMT_S16_LE | SND_PCM_FMT_U16_LE,	/* hardware formats */
        0,			/* align value */
        6,			/* minimal fragment */
        4000,			/* min. rate */
        48000,			/* max. rate */
        2,			/* max. voices */
        snd_es18xx_playback_open,
        snd_es18xx_playback_close,
        snd_es18xx_playback_ioctl,
        snd_es18xx_playback_prepare,
        snd_es18xx_playback_trigger,
        snd_es18xx_playback_pointer,
        snd_pcm1_playback_dma_ulaw,
        snd_pcm1_dma_move,
        snd_pcm1_playback_dma_neutral
};

static struct snd_stru_pcm1_hardware snd_es18xx_playback2 =
{
        NULL,			/* private data */
        NULL,			/* private_free */
        SND_PCM1_HW_AUTODMA,	/* flags */
        SND_PCM_FMT_MU_LAW |
        SND_PCM_FMT_U8 | SND_PCM_FMT_S8 | 
        SND_PCM_FMT_S16_LE | SND_PCM_FMT_U16_LE,	/* formats */
        SND_PCM_FMT_U8 | SND_PCM_FMT_S8 | 
        SND_PCM_FMT_S16_LE | SND_PCM_FMT_U16_LE,	/* hardware formats */
        0,			/* align value */
        6,			/* minimal fragment */
        4000,			/* min. rate */
        48000,			/* max. rate */
        2,			/* max. voices */
        snd_es18xx_playback2_open,
        snd_es18xx_playback2_close,
        snd_es18xx_playback2_ioctl,
        snd_es18xx_playback2_prepare,
        snd_es18xx_playback2_trigger,
        snd_es18xx_playback2_pointer,
        snd_pcm1_playback_dma_ulaw,
        snd_pcm1_dma_move,
        snd_pcm1_playback_dma_neutral
};

static struct snd_stru_pcm1_hardware snd_es18xx_record =
{
        NULL,			/* private data */
        NULL,			/* private free */
        SND_PCM1_HW_AUTODMA,	/* flags */
        SND_PCM_FMT_MU_LAW |
        SND_PCM_FMT_U8 | SND_PCM_FMT_S8 | 
        SND_PCM_FMT_S16_LE | SND_PCM_FMT_U16_LE,	/* formats */
        SND_PCM_FMT_U8 | SND_PCM_FMT_S8 | 
        SND_PCM_FMT_S16_LE | SND_PCM_FMT_U16_LE,	/* hardware formats */
        0,			/* align value */
        6,			/* minimal fragment */
        4000,			/* min. rate */
        48000,			/* max. rate */
        2,			/* max. voices */
        snd_es18xx_record_open,
        snd_es18xx_record_close,
        snd_es18xx_record_ioctl,
        snd_es18xx_record_prepare,
        snd_es18xx_record_trigger,
        snd_es18xx_record_pointer,
        snd_pcm1_record_dma_ulaw,
        snd_pcm1_dma_move,
        NULL
};

static void snd_es18xx_free(void *private_data)
{
}

snd_pcm_t *snd_es18xx_pcm_a(es18xx_t * codec)
{
        snd_pcm_t *pcm;
        snd_pcm1_t *pcm1;
	char str[16];
	if (codec->caps & ES18XX_PCM2)
		sprintf(str, "ES%x DAC2/ADC", codec->version);
	else
		sprintf(str, "ES%x", codec->version);
        pcm = snd_pcm1_new_device(codec->card, str);
        if (!pcm)
                return NULL;
        pcm1 = (snd_pcm1_t *) pcm->private_data;
        memcpy(&pcm1->playback.hw, &snd_es18xx_playback, sizeof(snd_es18xx_playback));
        memcpy(&pcm1->record.hw, &snd_es18xx_record, sizeof(snd_es18xx_record));
        pcm1->private_data = codec;
        pcm1->private_free = snd_es18xx_free;
        pcm->info_flags = SND_PCM_INFO_CODEC | SND_PCM_INFO_MMAP |
                SND_PCM_INFO_PLAYBACK | SND_PCM_INFO_RECORD |
                SND_PCM_INFO_DUPLEX;
	if (codec->caps & ES18XX_DUPLEX_SAME)
		pcm->info_flags |= SND_PCM_INFO_DUPLEX_LIMIT;
	if (codec->caps & ES18XX_DUPLEX_MONO)
		pcm->info_flags |= SND_PCM_INFO_DUPLEX_MONO;
	if (codec->caps & ES18XX_PCM2)
		sprintf(pcm->name, "ESS AudioDrive ES%x DAC2/ADC", codec->version);
	else
		sprintf(pcm->name, "ESS AudioDrive ES%x", codec->version);
        codec->pcm_a = pcm;
	return pcm;
}

snd_pcm_t *snd_es18xx_pcm_b(es18xx_t * codec)
{
        snd_pcm_t *pcm;
        snd_pcm1_t *pcm1;
	char str[16];
	sprintf(str, "ES%x DAC1", codec->version);
        pcm = snd_pcm1_new_device(codec->card, str);
        if (!pcm)
                return NULL;
        pcm1 = (snd_pcm1_t *) pcm->private_data;
        memcpy(&pcm1->playback.hw, &snd_es18xx_playback2, sizeof(snd_es18xx_playback2));
        pcm1->private_data = codec;
        pcm1->private_free = snd_es18xx_free;
        pcm->info_flags = SND_PCM_INFO_CODEC | SND_PCM_INFO_MMAP |
                SND_PCM_INFO_PLAYBACK;
        sprintf(pcm->name, "ESS AudioDrive ES%x DAC1", codec->version);
        codec->pcm_b = pcm;
	return pcm;
}


es18xx_t *snd_es18xx_new_device(snd_card_t * card,
                                unsigned short port,
                                unsigned short mpu_port,
                                unsigned short fm_port,
                                snd_irq_t * irqptr,
                                snd_dma_t * dma1ptr,
                                snd_dma_t * dma2ptr)
{
        es18xx_t *codec;

        codec = (es18xx_t *) snd_calloc(sizeof(es18xx_t));
        if (!codec)
                return NULL;
        snd_spin_prepare(codec, reg);
        snd_spin_prepare(codec, mixer);
        snd_spin_prepare(codec, ctrl);
        codec->card = card;
        codec->port = port;
        codec->mpu_port = mpu_port;
        codec->fm_port = fm_port;
        codec->irqptr = irqptr;
        codec->irq = irqptr->irq;
        codec->dma1ptr = dma1ptr;
        codec->dma1 = dma1ptr->dma;
        codec->dma2ptr = dma2ptr;
        codec->dma2 = dma2ptr->dma;
        codec->audio2_vol = 0x00;
	codec->open = 0;
	codec->active = 0;

        if (snd_es18xx_probe(codec) < 0) {
                snd_free(codec, sizeof(es18xx_t));
                return NULL;
        }
        return codec;
}

/*
 *  MIXER part
 */

static void snd_es18xx_record_volume_level(snd_kmixer_t * mixer,
					   snd_kmixer_channel_t * channel,
					   int l, int r)
{
        unsigned long flags;
        es18xx_t *codec;
        int left = (channel->record & SND_MIX_REC_LEFT) ? l : 0;
        int right = (channel->record & SND_MIX_REC_RIGHT) ? r : 0;
#if 0
	snd_printd("record volume %s (%02x,%02x)\n", channel->hw.name, left, right);
#endif
        codec = (es18xx_t *) mixer->private_data;
        snd_spin_lock(codec, mixer, &flags);
        switch (channel->hw.priority) {
        case SND_MIXER_PRI_PCM:
                snd_es18xx_mixer_write(codec, 0x69, left << 4 | right);
                break;
        case SND_MIXER_PRI_MIC:
                snd_es18xx_mixer_write(codec, 0x68, left << 4 | right);
                break;
        case SND_MIXER_PRI_FM:
                snd_es18xx_mixer_write(codec, 0x6b, left << 4 | right);
                break;
        case SND_MIXER_PRI_CD:
                snd_es18xx_mixer_write(codec, 0x6a, left << 4 | right);
                break;
        case SND_MIXER_PRI_AUXA:
                snd_es18xx_mixer_write(codec, 0x6c, left << 4 | right);
                break;
        case SND_MIXER_PRI_LINE:
                snd_es18xx_mixer_write(codec, 0x6e, left << 4 | right);
                break;
        default:
                snd_spin_unlock(codec, mixer, &flags);
                snd_printk("cannot set record volume for channel %s\n", channel->hw.name);
                return;
        }
        snd_spin_unlock(codec, mixer, &flags);
}


static void snd_es18xx_record_source(snd_kmixer_t * mixer,
				     snd_kmixer_channel_t *channel,
				     unsigned int rec)
{
        unsigned long flags;
        es18xx_t *codec;
        codec = (es18xx_t *) mixer->private_data;

	if (codec->caps & ES18XX_RECMIX) {
		int left = (rec & SND_MIX_REC_LEFT) ? channel->input.left : 0;
		int right = (rec & SND_MIX_REC_RIGHT) ? channel->input.right : 0;
#if 0
		snd_printd("record volume %s (%02x,%02x)\n", channel->hw.name, left, right);
#endif
		snd_spin_lock(codec, mixer, &flags);
		switch (channel->hw.priority) {
		case SND_MIXER_PRI_PCM:
			snd_es18xx_mixer_write(codec, 0x69, left << 4 | right);
			break;
		case SND_MIXER_PRI_MIC:
			snd_es18xx_mixer_write(codec, 0x68, left << 4 | right);
			break;
		case SND_MIXER_PRI_FM:
			snd_es18xx_mixer_write(codec, 0x6b, left << 4 | right);
			break;
		case SND_MIXER_PRI_CD:
			snd_es18xx_mixer_write(codec, 0x6a, left << 4 | right);
			break;
		case SND_MIXER_PRI_AUXA:
			snd_es18xx_mixer_write(codec, 0x6c, left << 4 | right);
			break;
		case SND_MIXER_PRI_LINE:
			snd_es18xx_mixer_write(codec, 0x6e, left << 4 | right);
			break;
		default:
			snd_spin_unlock(codec, mixer, &flags);
			snd_printk("cannot set record volume for channel %s\n", channel->hw.name);
			return;
		}
		snd_spin_unlock(codec, mixer, &flags);
	}
	else {
		enum { LINE=1, CD=2, MIC=4, MASTER=8 };
		int mask;
		switch (channel->hw.priority) {
		case SND_MIXER_PRI_LINE:
			mask = LINE;
			break;
		case SND_MIXER_PRI_CD:
			mask = CD;
			break;
		case SND_MIXER_PRI_MIC:
			mask = MIC;
			break;
		case SND_MIXER_PRI_MASTER:
			mask = MASTER;
			break;
		default:
			snd_printk("cannot set record source for channel %s\n", channel->hw.name);
			return;
		}
		if (rec & SND_MIX_REC)
			mixer->private_value |= mask;
		else
			mixer->private_value &= ~mask;

		snd_spin_lock(codec, mixer, &flags);

		switch (mixer->private_value) {
		case LINE:
			snd_es18xx_mixer_bits(codec, 0x1c, 0x06, 0x17);
			break;
		case CD:
			snd_es18xx_mixer_bits(codec, 0x1c, 0x02, 0x17);
			break;
		case MIC:
			snd_es18xx_mixer_bits(codec, 0x1c, 0x00, 0x17);
			break;
		case MASTER:
		default:
			snd_es18xx_mixer_bits(codec, 0x1c, 0x07, 0x17);
			break;
		}
		snd_spin_unlock(codec, mixer, &flags);
	}
}


static void snd_es18xx_volume_level(snd_kmixer_t * mixer,
                                    snd_kmixer_channel_t * channel,
                                    int left, int right)
{
        unsigned long flags;
        es18xx_t *codec;
	/* Hoping that right volume has the same flag set */
	if (left & SND_MIX_RECORD_VOLUME) {
		snd_es18xx_record_volume_level(mixer, channel,
					       SND_MIX_VOLUME(left),
					       SND_MIX_VOLUME(right));
		return;
	}
#if 0
	snd_printd("playback volume %s (%02x,%02x)\n", channel->hw.name, left, right);
#endif
        codec = (es18xx_t *) mixer->private_data;
        snd_spin_lock(codec, mixer, &flags);
        switch (channel->hw.priority) {
        case SND_MIXER_PRI_PCM:
		if (codec->caps & ES18XX_PCM2) {
			snd_es18xx_mixer_write(codec, 0x7c, (left << 4) | right);
			codec->audio2_vol = (left << 4) | right;
		}
		else
			snd_es18xx_mixer_write(codec, 0x14, (left << 4) | right);
                break;
        case SND_MIXER_PRI_PCM1:
                snd_es18xx_mixer_write(codec, 0x14, (left << 4) | right);
                break;
        case SND_MIXER_PRI_MIC:
                snd_es18xx_mixer_write(codec, 0x1a, (left << 4) | right);
                break;
        case SND_MIXER_PRI_MASTER:
                snd_es18xx_mixer_bits(codec, 0x60, left, 0x3f);
                snd_es18xx_mixer_bits(codec, 0x62, right, 0x3f);
                break;
        case SND_MIXER_PRI_FM:
                snd_es18xx_mixer_write(codec, 0x36, (left << 4) | right);
                break;
        case SND_MIXER_PRI_CD:
                snd_es18xx_mixer_write(codec, 0x38, (left << 4) | right);
                break;
        case SND_MIXER_PRI_AUXA:
                snd_es18xx_mixer_write(codec, 0x3a, (left << 4) | right);
                break;
        case SND_MIXER_PRI_SPEAKER:
                snd_es18xx_mixer_write(codec, 0x3c, left);
                break;
        case SND_MIXER_PRI_LINE:
                snd_es18xx_mixer_write(codec, 0x3e, (left << 4) | right);
                break;
        case SND_MIXER_PRI_GAIN:
                snd_es18xx_write(codec, 0xb4, (left << 4) | right);
                break;
        case SND_MIXER_PRI_3D_SPACE:
                snd_es18xx_mixer_write(codec, 0x52, left);
                break;
        default:
                snd_spin_unlock(codec, mixer, &flags);
		snd_printk("cannot set playback volume for channel %s\n", channel->hw.name);
                return;
        }
        snd_spin_unlock(codec, mixer, &flags);
}

static void snd_es18xx_mute(snd_kmixer_t * mixer, snd_kmixer_channel_t * channel, unsigned int mute)
{
        unsigned long flags;
        es18xx_t *codec;
#if 0
	snd_printd("mute %s (%02x)\n", channel->hw.name, mute);
#endif
        codec = (es18xx_t *) mixer->private_data;
        snd_spin_lock(codec, mixer, &flags);
        switch (channel->hw.priority) {
        case SND_MIXER_PRI_MASTER:
                snd_es18xx_mixer_bits(codec, 0x60, 
                                      (mute & SND_MIX_MUTE_LEFT) ? 0x40 : 0x00, 0x40);
                snd_es18xx_mixer_bits(codec, 0x62, 
                                      (mute & SND_MIX_MUTE_RIGHT) ? 0x40 : 0x00, 0x40);
                break;
        case SND_MIXER_PRI_3D_SPACE:
                snd_es18xx_mixer_bits(codec, 0x50, mute ? 0x00 : 0x0C, 0x0C);
                break;
        default:
                snd_spin_unlock(codec, mixer, &flags);
		snd_printk("cannot mute channel %s\n", channel->hw.name);
                return;
        }
        snd_spin_unlock(codec, mixer, &flags);
}


static struct snd_stru_mixer_channel_hw snd_es18xx_mixs[] =
{
        {
                SND_MIXER_PRI_PCM,	/* priority */
                SND_MIXER_PRI_MASTER,	/* parent priority */
                SND_MIXER_ID_PCM,	/* device name */
                SND_MIXER_OSS_PCM,	/* OSS device # */
                SND_MIXER_CINFO_CAP_STEREO | SND_MIXER_CINFO_CAP_RECORD | SND_MIXER_CINFO_CAP_RECORDVOLUME,
                0, 15,			/* min, max value */
                -4500, 0, 300,		/* min, max, step - dB */
                0,
                NULL,			/* compute dB -> linear */
                NULL,			/* compute linear -> dB */
                snd_es18xx_record_source, /* record source */
                NULL,			/* set mute */
                snd_es18xx_volume_level,/* set volume level */
                NULL,			/* set route */
        },
        {
                SND_MIXER_PRI_PCM1,	/* priority */
                SND_MIXER_PRI_MASTER,	/* parent priority */
                SND_MIXER_ID_PCM1,	/* device name */
                SND_MIXER_OSS_ALTPCM,	/* OSS device # */
                SND_MIXER_CINFO_CAP_STEREO,
                0, 15,			/* min, max value */
                -4500, 0, 300,		/* min, max, step - dB */
                0,
                NULL,			/* compute dB -> linear */
                NULL,			/* compute linear -> dB */
                NULL,			/* record source */
                NULL,			/* set mute */
                snd_es18xx_volume_level,/* set volume level */
                NULL,			/* set route */
        },
        {
                SND_MIXER_PRI_MIC,	/* priority */
                SND_MIXER_PRI_MASTER,	/* parent priority */
                SND_MIXER_ID_MIC,	/* device name */
                SND_MIXER_OSS_MIC,	/* OSS device # */
                SND_MIXER_CINFO_CAP_STEREO | SND_MIXER_CINFO_CAP_RECORD | SND_MIXER_CINFO_CAP_RECORDVOLUME,
                0, 15,			/* min, max value */
                -4500, 0, 300,		/* min, max, step - dB */
                0,
                NULL,			/* compute dB -> linear */
                NULL,			/* compute linear -> dB */
                snd_es18xx_record_source, /* record source */
                NULL,			/* set mute */
                snd_es18xx_volume_level,/* set volume level */
                NULL,			/* set route */
        },
        {
                SND_MIXER_PRI_MASTER,	/* priority */
                SND_MIXER_PRI_PARENT,	/* parent priority */
                SND_MIXER_ID_MASTER,	/* device name */
                SND_MIXER_OSS_VOLUME,	/* OSS device # */
                SND_MIXER_CINFO_CAP_STEREO | SND_MIXER_CINFO_CAP_RECORD,
                0, 63,			/* min, max value */
                -9450, 0, 150,		/* min, max, step - dB */
                0,
                NULL,			/* compute dB -> linear */
                NULL,			/* compute linear -> dB */
                snd_es18xx_record_source, /* record source */
                snd_es18xx_mute,	/* set mute */
                snd_es18xx_volume_level,/* set volume level */
                NULL,			/* set route */
        },
        {
                SND_MIXER_PRI_FM,	/* priority */
                SND_MIXER_PRI_MASTER,	/* parent priority */
                SND_MIXER_ID_FM,	/* device name */
                SND_MIXER_OSS_SYNTH,	/* OSS device # */
                SND_MIXER_CINFO_CAP_STEREO | SND_MIXER_CINFO_CAP_RECORD | SND_MIXER_CINFO_CAP_RECORDVOLUME,
                0, 15,			/* min, max value */
                -4500, 0, 300,		/* min, max, step - dB */
                0,
                NULL,			/* compute dB -> linear */
                NULL,			/* compute linear -> dB */
                snd_es18xx_record_source, /* record source */
                NULL,			/* set mute */
                snd_es18xx_volume_level,/* set volume level */
                NULL,			/* set route */
        },
        {
                SND_MIXER_PRI_CD,	/* priority */
                SND_MIXER_PRI_MASTER,	/* parent priority */
                SND_MIXER_ID_CD,	/* device name */
                SND_MIXER_OSS_CD,	/* OSS device # */
                SND_MIXER_CINFO_CAP_STEREO | SND_MIXER_CINFO_CAP_INPUT | SND_MIXER_CINFO_CAP_RECORD | SND_MIXER_CINFO_CAP_RECORDVOLUME,
                0, 15,			/* min, max value */
                -4500, 0, 300,		/* min, max, step - dB */
                0,
                NULL,			/* compute dB -> linear */
                NULL,			/* compute linear -> dB */
                snd_es18xx_record_source, /* record source */
                NULL,			/* set mute */
                snd_es18xx_volume_level,/* set volume level */
                NULL,			/* set route */
        },
        {
                SND_MIXER_PRI_AUXA,	/* priority */
                SND_MIXER_PRI_MASTER,	/* parent priority */
                SND_MIXER_ID_AUXA,	/* device name */
                SND_MIXER_OSS_LINE1,	/* OSS device # */
                SND_MIXER_CINFO_CAP_STEREO | SND_MIXER_CINFO_CAP_INPUT | SND_MIXER_CINFO_CAP_RECORD | SND_MIXER_CINFO_CAP_RECORDVOLUME,
                0, 15,			/* min, max value */
                -4500, 0, 300,		/* min, max, step - dB */
                0,
                NULL,			/* compute dB -> linear */
                NULL,			/* compute linear -> dB */
                snd_es18xx_record_source, /* record source */
                NULL,			/* set mute */
                snd_es18xx_volume_level,/* set volume level */
                NULL,			/* set route */
        },
        {
                SND_MIXER_PRI_SPEAKER,	/* priority */
                SND_MIXER_PRI_MASTER,	/* parent priority */
                SND_MIXER_ID_SPEAKER,	/* device name */
                SND_MIXER_OSS_SPEAKER,	/* OSS device # */
                SND_MIXER_CINFO_CAP_INPUT,
                0, 7,			/* max value */
                -2100, 0, 300,		/* min, max, step - dB */
                0,
                NULL,			/* compute dB -> linear */
                NULL,			/* compute linear -> dB */
                NULL,			/* record source */
                NULL,			/* set mute */
                snd_es18xx_volume_level,/* set volume level */
                NULL,			/* set route */
        },
        {
                SND_MIXER_PRI_LINE,	/* priority */
                SND_MIXER_PRI_MASTER,	/* parent priority */
                SND_MIXER_ID_LINE,	/* device name */
                SND_MIXER_OSS_LINE,	/* OSS device # */
                SND_MIXER_CINFO_CAP_STEREO | SND_MIXER_CINFO_CAP_INPUT | SND_MIXER_CINFO_CAP_RECORD | SND_MIXER_CINFO_CAP_RECORDVOLUME,
                0, 15,			/* min, max value */
                -4500, 0, 300,		/* min, max, step - dB */
                0,
                NULL,			/* compute dB -> linear */
                NULL,			/* compute linear -> dB */
                snd_es18xx_record_source, /* record source */
                NULL,			/* set mute */
                snd_es18xx_volume_level,/* set volume level */
                NULL,			/* set route */
        },
        {
                SND_MIXER_PRI_GAIN,	/* priority */
                SND_MIXER_PRI_PARENT,	/* parent priority */
                SND_MIXER_ID_GAIN,	/* device name */
                SND_MIXER_OSS_IMIX,	/* OSS device # */
                SND_MIXER_CINFO_CAP_STEREO,
                0, 15,			/* min, max value */
                -600, 1650, 150,	/* min, max, step - dB */
                0,
                NULL,			/* compute dB -> linear */
                NULL,			/* compute linear -> dB */
                NULL,			/* record source */
                NULL,			/* set mute */
                snd_es18xx_volume_level,/* set volume level */
                NULL,			/* set route */
        },
        {
                SND_MIXER_PRI_3D_SPACE,	/* priority */
                SND_MIXER_PRI_PARENT,	/* parent priority */
                SND_MIXER_ID_3D_SPACE,	/* device name */
                SND_MIXER_OSS_LINE3,	/* OSS device # */
                0,
                0, 63,			/* min, max value */
                0, 0, 0,		/* min, max, step - dB */
                0,
                NULL,			/* compute dB -> linear */
                NULL,			/* compute linear -> dB */
                NULL,			/* record source */
                snd_es18xx_mute,	/* set mute */
                snd_es18xx_volume_level,	/* set volume level */
                NULL,			/* set route */
        },
};

snd_kmixer_t *snd_es18xx_mixer(es18xx_t *codec)
{
        int idx;
        snd_kmixer_t *mixer;
        snd_kmixer_channel_t *channel;
	char str[16];

        if (!codec || !codec->card)
                return NULL;
	sprintf(str, "ES%x", codec->version);
        mixer = snd_mixer_new(codec->card, str);
        if (!mixer)
                return NULL;
        sprintf(mixer->name, "ESS AudioDrive ES%x", codec->version);

	if (!(codec->caps & ES18XX_RECMIX))
		mixer->hw.caps = SND_MIXER_INFO_CAP_EXCL_RECORD;
        mixer->private_data = codec;
	mixer->private_value = 0;

        for(idx = 0; idx < sizeof(snd_es18xx_mixs)/sizeof(struct snd_stru_mixer_channel_hw); idx++) {
		struct snd_stru_mixer_channel_hw * mix;
		mix = &snd_es18xx_mixs[idx];
		switch (mix->priority) {
		case SND_MIXER_PRI_MASTER:
			if (codec->caps & ES18XX_RECMIX)
				mix->caps &= ~SND_MIXER_CINFO_CAP_RECORD;
			break;
		case SND_MIXER_PRI_PCM:
		case SND_MIXER_PRI_FM:
			if (!(codec->caps & ES18XX_RECMIX))
				mix->caps &= ~SND_MIXER_CINFO_CAP_RECORD;
			break;
		case SND_MIXER_PRI_PCM1:
			if (!(codec->caps & ES18XX_PCM2))
				continue;
			break;
		case SND_MIXER_PRI_AUXA:
			if (!(codec->caps & ES18XX_AUXB))
				continue;
			break;
		case SND_MIXER_PRI_SPEAKER:
			if (!(codec->caps & ES18XX_SPEAKER))
				continue;
			break;
		case SND_MIXER_PRI_3D_SPACE:
			if (!(codec->caps & ES18XX_3D))
				continue;
			break;
		}
		if (mix->caps & SND_MIXER_CINFO_CAP_RECORD) {
			if (!(codec->caps & ES18XX_RECMIX)) {
				mix->caps &= ~SND_MIXER_CINFO_CAP_RECORDVOLUME;
				mix->caps |= SND_MIXER_CINFO_CAP_JOINRECORD;
			}
		}
		else {
			mix->caps &= ~(SND_MIXER_CINFO_CAP_JOINRECORD |
				       SND_MIXER_CINFO_CAP_RECORDVOLUME);
			mix->set_record_source = 0;
		}
		channel = snd_mixer_new_channel(mixer, mix);
                if (!channel) {
                        snd_mixer_free(mixer);
                        return NULL;
                }
        }
        return mixer;
}

/*
 *  INIT part
 */

#ifndef LINUX_2_1
extern struct symbol_table snd_symbol_table_es18xx_export;
#endif

int init_module(void)
{
#ifndef LINUX_2_1
        if (register_symtab(&snd_symbol_table_es18xx_export) < 0)
                return -ENOMEM;
#endif
        return 0;
}

void cleanup_module(void)
{
}

